//*** LICENSE ***
//ColumnListView, its associated classes and source code, and the other components of Santa's Gift Bag are
//being made publicly available and free to use in freeware and shareware products with a price under $25
//(I believe that shareware should be cheap). For overpriced shareware (hehehe) or commercial products,
//please contact me to negotiate a fee for use. After all, I did work hard on this class and invested a lot
//of time into it. That being said, DON'T WORRY I don't want much. It totally depends on the sort of project
//you're working on and how much you expect to make off it. If someone makes money off my work, I'd like to
//get at least a little something.  If any of the components of Santa's Gift Bag are is used in a shareware
//or commercial product, I get a free copy.  The source is made available so that you can improve and extend
//it as you need. In general it is best to customize your ColumnListView through inheritance, so that you
//can take advantage of enhancements and bug fixes as they become available. Feel free to distribute the 
//ColumnListView source, including modified versions, but keep this documentation and license with it.

//*** DESCRIPTION ***
//Class which buffers file access.   Lacks the extra features provided by BufferedFile, but this is useful for
//buffering output to a text file or other file formats where the extra byte that BufferedFile sticks on the
//front of the file to indicate endianness interferes with the output file.

//BufferedFile still provides all of the functionality of a BFile: It now provides Read/Write access, random
//access, etc.  All the regular BFile calls can be mixed with the BufferedFile calls, although the BFile calls
//won't benefit from the buffering (they force the buffer to flush, and actually introduce inefficiency - use
//them carefully; in general try to stick to the BufferedRead and BufferedWrite calls)


//******************************************************************************************************
//**** SYSTEM HEADER FILES
//******************************************************************************************************
#include <stdio.h>


//******************************************************************************************************
//**** PROJECT HEADER FILES
//******************************************************************************************************
#include "BufferedFile.h"


//******************************************************************************************************
//**** BufferedFile CLASS DEFINITION
//******************************************************************************************************
BufferedFile::BufferedFile(const entry_ref* ref, uint32 open_Mode, int32 Buffer_Size)
: BFile(ref,open_Mode)
{
	openMode = open_Mode;
	Valid = BFile::InitCheck();
	AllocAndInitBuffer(Buffer_Size);
}


BufferedFile::BufferedFile(const BEntry* entry, uint32 open_Mode, int32 Buffer_Size)
: BFile(entry,open_Mode)
{
	openMode = open_Mode;
	Valid = BFile::InitCheck();
	AllocAndInitBuffer(Buffer_Size);
}


BufferedFile::BufferedFile(const char* path, uint32 open_Mode, int32 Buffer_Size)
: BFile(path,open_Mode)
{
	openMode = open_Mode;
	Valid = BFile::InitCheck();
	AllocAndInitBuffer(Buffer_Size);
}


BufferedFile::BufferedFile(BDirectory* dir, const char* path, uint32 open_Mode, int32 Buffer_Size)
: BFile(dir,path,open_Mode)
{
	openMode = open_Mode;
	Valid = BFile::InitCheck();
	AllocAndInitBuffer(Buffer_Size);
}


void BufferedFile::AllocAndInitBuffer(int32 Buffer_Size)
{
	Buffer = NULL;
	if(Valid == B_NO_ERROR)
	{
		//Allocate the buffer
		BufferSize = Buffer_Size;
		Buffer = new char[BufferSize];
		BytesInBuffer = 0;
		BufferPos = 0;
		WriteBuffered = false;
	}
}


BufferedFile::~BufferedFile()
{
	if(Buffer)
		delete[] Buffer;
}


status_t BufferedFile::FlushBuffer()
//Flushes the read buffer and moves the file pointer to the position that was being read from.  It is
//automatically called when necessary when working within the file, but if you have buffered write data, call
//it yourself before deleting the BufferedFile to make sure the data is written successfully (the destructor
//can't return an error code).
{
	if(WriteBuffered)
	{
		if(BFile::Write(Buffer,BufferPos) != BufferPos)
			return B_FILE_ERROR;
		BufferPos = 0;
		WriteBuffered = false;
	}
	else if(BufferPos != 0)
	{
		off_t Result = BFile::Seek(off_t(BufferPos)-off_t(BytesInBuffer),SEEK_CUR);
		if(Result == B_ERROR || Result == B_FILE_ERROR)
			return B_FILE_ERROR;
		BufferPos = 0;
	}
	return B_NO_ERROR;
}


status_t BufferedFile::InitCheck()
//Returns B_NO_ERROR if the BufferedFile was created successfully.  Returns B_NO_INIT if initialization of
//the underlying BFile failed.  Returns B_ERROR if reading or writing the endianness flag to or from the
//first byte of the file failed.
{
	return Valid;
}


ssize_t BufferedFile::BufferedWrite(const void* WriteBuffer, size_t BytesToWrite)
//Writes the raw data contained in WriteBuffer.  Returns the number of bytes actually written.
{
	if(!IsWritable())
		return 0;
	ssize_t Counter = 0;
	if(!WriteBuffered && BufferPos != 0)
		if(FlushBuffer() != B_NO_ERROR)
			return 0;
	while(Counter < (ssize_t)BytesToWrite)
	{
		Buffer[BufferPos++] = ((const char*)WriteBuffer)[Counter++];
		if(BufferPos == BufferSize)
		{
			//Buffer is full, need to write and reset
			int32 Result = BFile::Write(Buffer,BufferSize);
			if(Result != BufferSize)
			{
				WriteBuffered = false;
				return Counter-(BufferSize-Result);
			}
			BufferPos = 0;
		}
	}
	if(BufferPos > 0)
		WriteBuffered = true;
	return Counter;
}


ssize_t BufferedFile::BufferedRead(void* ReadBuffer, size_t BytesToRead)
//Reads raw data into ReadBuffer.  Returns the number of bytes actually read.
{
	if(!IsReadable())
		return 0;
	if(WriteBuffered)
		if(FlushBuffer() != B_NO_ERROR)
			return 0;
	ssize_t Counter = 0;
	while(Counter < (ssize_t)BytesToRead)
	{
		if(BufferPos == 0)
		{
			ssize_t Result = BFile::Read(Buffer,BufferSize);
			BytesInBuffer = Result;
			if(Result < ssize_t(BufferSize))
			{
				if(Result <= 0)
					return Counter;
			}
		}
		((char*)ReadBuffer)[Counter++] = ((const char*)Buffer)[BufferPos++];
		if(BufferPos == BytesInBuffer)
		{
			//Reached end of buffer, need to reset for another read
			if(BytesInBuffer != BufferSize)
			{
				//Already read to the end of the file
				BufferPos = 0;
				BytesInBuffer = 0;
				return Counter;
			}
			BufferPos = 0;
			BytesInBuffer = 0;
		}
	}
	return Counter;
}


ssize_t BufferedFile::BufferedPreview(void* PreviewBuffer,  size_t BytesToPreview)
{
	if(BytesToPreview > (size_t)BufferSize)
		return 0;
	if(BytesToPreview > ((size_t)(BytesInBuffer-BufferPos)) && BufferPos > 0)
	{
		//Need to shift the buffer and load more
		int32 BytesToShift = BytesInBuffer-BufferPos;
		for(int32 NewPos = 0; NewPos < BytesToShift; NewPos++)
			Buffer[NewPos] = Buffer[BufferPos+NewPos];
		BytesInBuffer -= BufferPos;
		BufferPos = 0;
		ssize_t Result = BFile::Read(&Buffer[BytesInBuffer],BufferSize-BytesInBuffer);
		BytesInBuffer += Result;
	}

	ssize_t Counter = 0;
	int32 PreviewPos = BufferPos;
	while(Counter < (ssize_t)BytesToPreview && PreviewPos < BytesInBuffer)
		((char*)PreviewBuffer)[Counter++] = Buffer[PreviewPos++];

	return Counter;
}


ssize_t BufferedFile::BufferedReadLine(void* LineBuffer, size_t LineBufferSize)
{
	int32 LineBufferPos = 0;
	ssize_t TotalBytesRead = 0;
	if(LineBufferSize == 0)
		return 0;
	while(true)
	{
		ssize_t BytesRead = BufferedRead(&((char*)LineBuffer)[LineBufferPos],1);
		TotalBytesRead += BytesRead;
		if(BytesRead == 0)
		{
			((char*)LineBuffer)[LineBufferPos] = 0;
			return TotalBytesRead;
		}
		if(((char*)LineBuffer)[LineBufferPos] == '\r')
		{
			char PreviewBuffer;
			if(BufferedPreview(&PreviewBuffer,1) == 1 && PreviewBuffer == '\n')
				TotalBytesRead += BufferedRead(&PreviewBuffer,1);
			((char*)LineBuffer)[LineBufferPos] = 0;
			return TotalBytesRead;
		}
		if(((char*)LineBuffer)[LineBufferPos] == '\n')
		{
			((char*)LineBuffer)[LineBufferPos] = 0;
			return TotalBytesRead;
		}
		if(((char*)LineBuffer)[LineBufferPos] == 0)
			return TotalBytesRead;
		LineBufferPos++;
		if(LineBufferPos == (int32)(LineBufferSize-1))
		{
			((char*)LineBuffer)[LineBufferPos] = 0;
			return TotalBytesRead;
		}
	}
}


ssize_t BufferedFile::Read(void *buffer, size_t size)
{
	if(FlushBuffer() != B_NO_ERROR)
		return 0;
	return BFile::Read(buffer,size);
}


ssize_t BufferedFile::ReadAt(off_t pos, void *buffer, size_t size)
{
	if(FlushBuffer() != B_NO_ERROR)
		return 0;
	return BFile::ReadAt(pos,buffer,size);
}


ssize_t BufferedFile::Write(const void *buffer, size_t size)
{
	if(FlushBuffer() != B_NO_ERROR)
		return 0;
	return BFile::Write(buffer,size);
}


ssize_t BufferedFile::WriteAt(off_t pos, const void *buffer, size_t size)
{
	if(FlushBuffer() != B_NO_ERROR)
		return 0;
	return BFile::WriteAt(pos,buffer,size);
}


off_t BufferedFile::Seek(off_t position, uint32 seek_mode)
{
	if(FlushBuffer() != B_NO_ERROR)
		return 0;
	return BFile::Seek(position,seek_mode);
}


status_t BufferedFile::SetSize(off_t size)
{
	if(FlushBuffer() != B_NO_ERROR)
		return B_FILE_ERROR;
	return BFile::SetSize(size);
}


status_t BufferedFile::SetTo(const entry_ref *ref, uint32 open_mode)
{
	if(FlushBuffer() != B_NO_ERROR)
		return B_FILE_ERROR;
	Valid = BFile::SetTo(ref,open_mode);
	openMode = open_mode;
	if(Valid != B_NO_ERROR)
		return Valid;
	Valid = BFile::InitCheck();
	return Valid;
}


status_t BufferedFile::SetTo(const BEntry *entry, uint32 open_mode)
{
	if(FlushBuffer() != B_NO_ERROR)
		return B_FILE_ERROR;
	Valid = BFile::SetTo(entry,open_mode);
	openMode = open_mode;
	if(Valid != B_NO_ERROR)
		return Valid;
	Valid = BFile::InitCheck();
	return Valid;
}


status_t BufferedFile::SetTo(const char *path, uint32 open_mode)
{
	if(FlushBuffer() != B_NO_ERROR)
		return B_FILE_ERROR;
	Valid = BFile::SetTo(path,open_mode);
	openMode = open_mode;
	if(Valid != B_NO_ERROR)
		return Valid;
	Valid = BFile::InitCheck();
	return Valid;
}


status_t BufferedFile::SetTo(const BDirectory *dir, const char *path, uint32 open_mode)
{
	if(FlushBuffer() != B_NO_ERROR)
		return B_FILE_ERROR;
	Valid = BFile::SetTo(dir,path,open_mode);
	openMode = open_mode;
	if(Valid != B_NO_ERROR)
		return Valid;
	Valid = BFile::InitCheck();
	return Valid;
}
