//Name:		ChecksumDataBuffer.cpp
//Author:	Brian Tietz
//Description:
//	A buffer for writing data to for network transmission.  The buffer grows as needed to hold the
//  data as it's added, and can return a checksum.  The recommended checksum method is to use a 48-bit
//  checksum by sending both GetChecksum16() and GetWrappingChecksum().  GetChecksum just adds each byte
//  of the data buffer to a uint16 or uint32, depending on the function version.  GetWrappingChecksum
//  generates the sum for all bytes in the data buffer of (byte_value<<(pos%25)) in a uint32.  On the
//  receiving end, the static versions of the checksum functions can be used to generate a checksum
//  on read data by the same algorithm.
//Conventions:
//	Global constants (declared with const) and #defines - begin with "c_" followed by lowercase
//		words separated by underscores.
//		(E.G., #define c_my_constant 5).
//		(E.G., const int c_my_constant = 5;).
//	Global variables - begin with "g_" followed by lowercase words separated by underscores.
//		(E.G., int g_my_global;).
//	New data types (classes, structs, typedefs, etc.) - begin with an uppercase letter followed by
//		lowercase words separated by uppercase letters.  Enumerated constants contain a prefix
//		associating them with a particular enumerated set.
//		(E.G., typedef int MyTypedef;).
//		(E.G., enum MyEnumConst {c_mec_one, c_mec_two};)
//	Private member variables - begin with "m_" followed by lowercase words separated by underscores.
//		(E.G., int m_my_member;).
//	Public or friend-accessible member variables - all lowercase words separated by underscores.
//		(E.G., int public_member;).
//	Argument and local variables - begin with a lowercase letter followed by
//		lowercase words separated by underscores.  If the name is already taken by a public member
//		variable, prefix with a_ or l_
//		(E.G., int my_local; int a_my_arg, int l_my_local).
//	Functions (member or global) - begin with an uppercase letter followed by lowercase words
//		separated by uppercase letters.
//		(E.G., void MyFunction(void);).


//******************************************************************************************************
//**** Project header files
//******************************************************************************************************
#include "ChecksumDataBuffer.h"


//******************************************************************************************************
//**** ChecksumDataBuffer
//******************************************************************************************************
ChecksumDataBuffer::ChecksumDataBuffer(uint32 growth_increment, bool use_network_order)
{
	m_data = new uint8[growth_increment];
	m_growth_increment = growth_increment;;
	m_buffer_size = growth_increment;
	m_buffer_pos = 0;
	m_read_pos = 0;
	m_use_network_order = use_network_order;
}


ChecksumDataBuffer::~ChecksumDataBuffer()
{
	delete[] m_data;
}


void ChecksumDataBuffer::AddNumber(void *data, uint8 size)
{
	uint32 i;
	if(m_buffer_pos + size > m_buffer_size)
	{
		m_buffer_size = ((m_buffer_pos+size+(m_growth_increment-1))/m_growth_increment)*
			m_growth_increment;
		uint8* new_buffer = new uint8[m_buffer_size];
		for(i=0; i<m_buffer_pos; i++)
			new_buffer[i] = m_data[i];
		delete[] m_data;
		m_data = new_buffer;
	}

	#ifdef __INTEL__
		if(m_use_network_order)
			for(i=0; i<size; i++)
				m_data[m_buffer_pos+i] = ((uint8*)data)[size-1-i];
		else
			for(i=0; i<size; i++)
				m_data[m_buffer_pos+i] = ((uint8*)data)[i];
	#else
		for(i=0; i<size; i++)
			m_data[m_buffer_pos+i] = ((uint8*)data)[i];
	#endif
	m_buffer_pos += size;
}


void ChecksumDataBuffer::AddData(const void *data, uint32 length)
{
	uint32 i;
	if(m_buffer_pos + length > m_buffer_size)
	{
		m_buffer_size = ((m_buffer_pos+length+(m_growth_increment-1))/m_growth_increment)*
			m_growth_increment;
		uint8* new_buffer = new uint8[m_buffer_size];
		for(i=0; i<m_buffer_pos; i++)
			new_buffer[i] = m_data[i];
		delete[] m_data;
		m_data = new_buffer;
	}

	for(i=0; i<length; i++)
	{
		m_data[m_buffer_pos] = ((uint8*)data)[i];
		m_buffer_pos++;
	}
}


void ChecksumDataBuffer::ExpandBufferForAdd(uint32 length)
{
	uint32 i;
	if(m_buffer_pos + length > m_buffer_size)
	{
		m_buffer_size = ((m_buffer_pos+length+(m_growth_increment-1))/m_growth_increment)*
			m_growth_increment;
		uint8* new_buffer = new uint8[m_buffer_size];
		for(i=0; i<m_buffer_pos; i++)
			new_buffer[i] = m_data[i];
		delete[] m_data;
		m_data = new_buffer;
	}
}


void ChecksumDataBuffer::ReplaceNumber(void *data, uint8 size, uint32 at_pos)
{
	uint8 i;
	#ifdef __INTEL__
		if(m_use_network_order)
			for(i=0; i<size; i++)
				m_data[at_pos+i] = ((uint8*)data)[size-1-i];
		else
			for(i=0; i<size; i++)
				m_data[at_pos+i] = ((uint8*)data)[i];
	#else
		for(i=0; i<size; i++)
			m_data[at_pos+i] = ((uint8*)data)[i];
	#endif
}


void ChecksumDataBuffer::ReplaceData(const void *data, uint32 length, uint32 at_pos)
{
	for(uint32 i=0; i<length; i++)
		m_data[at_pos+i] = ((uint8*)data)[i];
}


uint16 ChecksumDataBuffer::GetChecksum16(uint8 *data, uint32 length)
{
	uint16 checksum = 0;
	for(uint32 i=0; i<length; i++)
		checksum += data[i];
	return checksum;
}


uint32 ChecksumDataBuffer::GetChecksum32(uint8 *data, uint32 length)
{
	uint32 checksum = 0;
	for(uint32 i=0; i<length; i++)
		checksum += data[i];
	return checksum;
}


uint32 ChecksumDataBuffer::GetWrappingChecksum(uint8 *data, uint32 length)
{
	uint32 checksum = 0;
	for(uint32 i=0; i<length; i++)
		checksum += (((uint32)data[i])<<(i%25));
	return checksum;
}


uint32 ChecksumDataBuffer::AddMoreWrappingChecksum(uint8 *data, uint32 length, uint32 prev_data_length,
	uint32 prev_checksum)
{
	uint32 checksum = prev_checksum;
	for(uint32 i=0; i<length; i++)
		checksum += (((uint32)data[i])<<((i+prev_data_length)%25));
	return checksum;
}


void ChecksumDataBuffer::ReadNumber(void* data, uint8 size)
{
	uint8 i;
	#ifdef __INTEL__
		if(m_use_network_order)
			for(i=0; i<size; i++)
				((uint8*)data)[size-1-i] = m_data[m_read_pos+i];
		else
			for(i=0; i<size; i++)
				((uint8*)data)[i] = m_data[m_read_pos+i];
	#else
		for(i=0; i<size; i++)
			((uint8*)data)[i] = m_data[m_read_pos+i];
	#endif
	m_read_pos+=size;
}


void ChecksumDataBuffer::ReadNumberAt(void *data, uint8 size, uint32 at_pos) const
{
	uint8 i;
	#ifdef __INTEL__
		if(m_use_network_order)
			for(i=0; i<size; i++)
				((uint8*)data)[size-1-i] = m_data[at_pos+i];
		else
			for(i=0; i<size; i++)
				((uint8*)data)[i] = m_data[at_pos+i];
	#else
		for(i=0; i<size; i++)
			((uint8*)data)[i] = m_data[at_pos+i];
	#endif
}


